/**
 * ***************************************************************************** Institute for
 * System Programming of the Russian Academy of Sciences Copyright (c) 2009-2013 All rights
 * reserved. This program and the accompanying materials are made available under the terms of the
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * ****************************************************************************
 */
package com.unitesk.requality_git;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.egit.core.GitProvider;
import org.eclipse.jgit.api.*;
import org.eclipse.jgit.lib.AnyObjectId;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectLoader;
import org.eclipse.jgit.lib.Ref;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevTree;
import org.eclipse.jgit.revwalk.RevWalk;
import org.eclipse.jgit.treewalk.TreeWalk;
import org.eclipse.jgit.treewalk.filter.PathFilter;
import org.eclipse.team.core.RepositoryProvider;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;

import com.unitesk.requality.core.FileSpaceTimePoint;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.eclipse.cvs.CVSException;
import com.unitesk.requality.eclipse.cvs.ICVSClient;

public class SVCGITClient implements ICVSClient {

  public static final String ID = "Subclipse.GITClientAdapter";
  IProject project = null;
  Repository repo = null;
  Git jGit = null;
  GitProvider teamProvider = null;
  Ref revHEAD = null;
  String currRev = null;
  String repoURL = null;

  @Override
  public String getID() {
    return ID;
  }

  // Конструкторы и их подметоды
  public SVCGITClient() throws Exception {}

  public SVCGITClient(IProject proj) throws Exception {
    if (proj != null) init(proj);
  }

  // Методы, определённые в интерфейсе
  @Override
  public String getFileContentFromRepoS(
      String filePathInRepo, String pegRev, String revOfIntrst, IProgressMonitor monitor)
      throws Exception {
    // no realization of peg revisions in GIT;
    return getFileContentFromRepoS(filePathInRepo, revOfIntrst, monitor);
  }

  @Override
  public String getFileContentFromRepoS(String filePathInRepo, String rev, IProgressMonitor monitor)
      throws Exception {

    if (jGit == null) throw new Exception("JGit client has not been initialised");

    if (monitor != null) {
      monitor.beginTask("Getting file " + filePathInRepo, 1);
    }

    ObjectId revID = repo.resolve(rev);
    RevWalk revWalk = new RevWalk(repo);
    RevCommit commit = revWalk.parseCommit(revID);
    RevTree tree = commit.getTree();
    TreeWalk treeWalk = new TreeWalk(repo);
    treeWalk.addTree(tree);
    treeWalk.setRecursive(true);
    treeWalk.setFilter(PathFilter.create(filePathInRepo));
    try{
      if (!treeWalk.next()) {
        throw new Exception("File not found");
      }
  
      ObjectId objectId = treeWalk.getObjectId(0);
      ObjectLoader loader = repo.open(objectId);
  
      ByteArrayOutputStream out = new ByteArrayOutputStream();
      loader.copyTo(out);
      if (monitor != null) {
        monitor.worked(1);
        monitor.done();
      }
      return out.toString();
    }
    finally{
      revWalk.close();
      treeWalk.close();
    }
  }

  @Override
  public File getContainerFromRepoS(String filePathInRepo, String rev, IProgressMonitor monitor)
      throws Exception {
    long k = System.currentTimeMillis();
    if (repoURL == null || repoURL.isEmpty()) {
      throw new Exception("getFileFromRepoS: repository URL is null or empty");
    }
    if (filePathInRepo == null) {
      throw new Exception("getFileFromRepoS: filePathInRepo is null");
    }
    RLogger.logInfo("start: " + ((System.currentTimeMillis()) - k));
    k = System.currentTimeMillis();

    if (monitor != null) {
      monitor.beginTask("Getting file " + filePathInRepo, 1);
    }

    // ревизия файла
    jGit.log().all();
    jGit.log().addPath(filePathInRepo);

    File par = getRevisionDirectory(rev);

    RLogger.logInfo("precheck: " + ((System.currentTimeMillis()) - k));
    k = System.currentTimeMillis();
//
    Git.cloneRepository().setDirectory(par).setURI(repoURL).call().checkout().setName(rev).call();

    RLogger.logInfo("checkout: " + ((System.currentTimeMillis()) - k));
    k = System.currentTimeMillis();

    markDelFile(par.getParentFile().getParentFile());
    RLogger.logInfo(
        "markdel: " + par.getParentFile().getParentFile() + " " + ((System.currentTimeMillis()) - k));
    k = System.currentTimeMillis();

    return new File(par.getAbsolutePath(), filePathInRepo);
  }

  private void markDelFile(File f) {
    if (f.isFile()) f.deleteOnExit();
    else {
      f.deleteOnExit();
      // for(File f2:f.listFiles())
      // markDelFile(f2);
    }
  }

  File tempParent = null;

  private File getRevisionDirectory(String rev) throws IOException {
    if (workingPlace == null || !workingPlace.exists()) {
      tempParent = File.createTempFile("requality", "gitRevisions");
      tempParent = new File(tempParent.getAbsolutePath() + "_content");
      tempParent.mkdir();
    } else tempParent = workingPlace;
    File newDir =
        new File(
            tempParent.getAbsolutePath() + File.separator + "cache" + File.separator + "trunk",
            rev);
    newDir = new File(newDir.getAbsolutePath(), "git");
    if (!newDir.exists()) newDir.mkdirs();
    return newDir;
  }

  @Override
  public ArrayList<FileSpaceTimePoint> getResourceRevisions(
      IResource componentFile, IProgressMonitor monitor) throws Exception {
    return getResourceRevisions(componentFile, null, monitor);
  }

  @Override
  public ArrayList<FileSpaceTimePoint> getResourceRevisions(
      IResource componentFile, String startRev, IProgressMonitor monitor) throws Exception {
    if (monitor == null) monitor = new NullProgressMonitor();

    monitor.beginTask("Getting file revision list", 2);

    if (componentFile == null) {
      monitor.done();
      throw new Exception("Component file link is null");
    }

    String path =
        teamProvider.getData().getRepositoryMapping(project).getRepoRelativePath(componentFile);

    if (path == null) {
      monitor.done();
      throw new Exception("Could not find relative path of the file");
    }

    jGit.log().all();
    jGit.log().addPath(path);

    if (startRev != null && !startRev.isEmpty()) {
      AnyObjectId start = ObjectId.fromString(startRev);
      jGit.log().add(start);
    }

    Iterable<RevCommit> itleRC = jGit.log().call();
    Iterator<RevCommit> itRC = itleRC.iterator();
    if (!itRC.hasNext()) {
      monitor.done();
      return null;
    }
    RevCommit rcommit = null;

    monitor.worked(1);
    if (monitor.isCanceled()) {
      monitor.done();
      throw new InterruptedException();
    }

    ArrayList<FileSpaceTimePoint> result = new ArrayList<FileSpaceTimePoint>();
    FileSpaceTimePoint point = null;

    while (itRC.hasNext()) {
      if (monitor.isCanceled()) {
        monitor.done();
        throw new InterruptedException();
      }
      point = new FileSpaceTimePoint();
      rcommit = itRC.next();

      RevTree tree = rcommit.getTree();
      TreeWalk treeWalk = new TreeWalk(repo);
      try{
        treeWalk.addTree(tree);
        treeWalk.setRecursive(true);
        treeWalk.setFilter(PathFilter.create(path));
        if (treeWalk.next()) {
          point.setPath(path);
          point.setRevision(rcommit.getName());
          point.setEpochTimeSec(rcommit.getCommitTime());
          point.setMessage(rcommit.getShortMessage() + " (" + rcommit.getFullMessage() + ")");
          point.setAuthor(rcommit.getCommitterIdent().getName());
  
          result.add(0, point);
        }
        point = null;
      }
      finally{
        treeWalk.close();
      }
    }

    monitor.worked(1);
    monitor.done();
    return result;
  }

  private File workingPlace;

  @Override
  public void init(IProject proj) throws Exception {
    if (proj == null) throw new Exception("NULL project");
    project = proj;
    RepositoryProvider tProv = RepositoryProvider.getProvider(project);

    if (!(tProv instanceof GitProvider))
      throw new Exception("Team provider seems to be NOT EGit, but should BE EGit");
    teamProvider = (GitProvider) tProv;

    String projectAbsolutePath = project.getLocationURI().getPath().toString() + File.separator;
    workingPlace = new java.io.File(projectAbsolutePath + "/.reqcvs/");

    if (!workingPlace.exists())
      if (!workingPlace.mkdir()) {
        throw new CVSException("Cannot create folder " + workingPlace.getAbsolutePath());
      }

    repo = teamProvider.getData().getRepositoryMapping(project).getRepository();
    //TODO path
    repoURL =
        teamProvider
            .getData()
            .getRepositoryMapping(project)
            .getGitDirAbsolutePath()
            .toFile()
            .getParent()
            .toString();
    jGit = new Git(repo);

    if (repoURL == null || repoURL.length() == 0)
      throw new Exception("Empty repository URL string");
    if (repoURL.charAt(repoURL.length() - 1) == '/')
      repoURL = repoURL.substring(0, repoURL.length() - 1);
  }

  public void setCurrentRevision(String rev) throws Exception {
    currRev = rev;
  }

  @Override
  public String getLocalCopyRevision(IResource componentFile) throws Exception {
    if (componentFile == null) return null;
    String str =
        project.getWorkspace().getRoot().getLocation().toFile().toString()
            + componentFile.getFullPath().toString();
    str = str.replace(IPath.SEPARATOR, File.separatorChar);
    int i = 0;
    while (str.charAt(i) == '/' || str.charAt(i) == '\\') i++;
    str = str.substring(i);

    Iterable<RevCommit> logrecs = jGit.log().addPath(str).call();
    Iterator<RevCommit> logrec = logrecs.iterator();

    if (logrec.hasNext()) return logrec.next().getName();
    else return null;
  }

  @Override
  public void commit(IFile file, String commitMessage, IProgressMonitor monitor) throws Exception {
    if (monitor == null) monitor = new NullProgressMonitor();

    monitor.beginTask("Committing changes", 1);
    // добавить файл в index для наката
    String str = file.getFullPath().toString();

    int i = 0;
    while (str.charAt(i) == '/' || str.charAt(i) == '\\') i++;
    str = str.substring(i);
    str = str.replaceFirst(project.getName(), "");
    i = 0;
    while (str.charAt(i) == '/' || str.charAt(i) == '\\') i++;
    str = str.substring(i);

    jGit.add().addFilepattern(str).call();

    if (monitor.isCanceled()) {
      monitor.done();
      throw new InterruptedException();
    }
    // накатить файл
    jGit.commit().setMessage(commitMessage).call();
    monitor.worked(1);
    monitor.done();
    // throw new Exception("Not implemented yet");

  }

  @Override
  public String getRepoURL() {
    return repoURL;
  }

  @Override
  public String[] getRelativePathAndLatestRevForAFile(IFile componentFile) throws Exception {
    if (componentFile == null) throw new Exception("Component file link is null");

    String path =
        teamProvider.getData().getRepositoryMapping(project).getRepoRelativePath(componentFile);

    if (path == null) throw new Exception("Could not find relative path of the file");

    Iterable<RevCommit> itleRC = jGit.log().addPath(path).call();

    Iterator<RevCommit> itRC = itleRC.iterator();

    int epochTime = 0;
    String revName = null;
    RevCommit rcommit = null;

    if (itRC.hasNext()) {
      rcommit = itRC.next();

      revName = rcommit.getName();
      epochTime = rcommit.getCommitTime();

      while (itRC.hasNext()) {
        rcommit = itRC.next();

        if (rcommit.getCommitTime() > epochTime) {
          revName = rcommit.getName();
          epochTime = rcommit.getCommitTime();
        }
      }

      String res[] = new String[3];
      res[0] = path;
      res[1] = revName;
      res[2] = String.valueOf(epochTime);

      return res;
    }

    return null;
  }

  // вспомогательные методы
  @Override
  public void excludeDotReqCSVDir() throws Exception {
    if (this.jGit == null || this.project == null) {
      throw new Exception("The CVCL SVC client has not been properly / seccessfully initialized");
    }

    File gitIgnoreFile =
        new File(project.getLocationURI().getPath().toString() + File.separator + ".gitignore");

    if (!gitIgnoreFile.exists()) {
      gitIgnoreFile.createNewFile();
    }

    BufferedReader reader = new BufferedReader(new FileReader(gitIgnoreFile));
    String str = reader.readLine();
    while (str != null) {
      if (str.trim().equalsIgnoreCase(".cvcl/")) {
        reader.close();
        return;
      }
      str = reader.readLine();
    }
    reader.close();

    int c = '\n';

    if (gitIgnoreFile.length() > 0) {
      reader = new BufferedReader(new FileReader(gitIgnoreFile));
      reader.skip(gitIgnoreFile.length() - 1);
      c = reader.read();
      reader.close();
    }

    BufferedWriter writer = new BufferedWriter(new FileWriter(gitIgnoreFile, true));
    if (c != '\n' && c != '\r') writer.newLine();
    writer.write(".cvcl/");
    writer.newLine();
    writer.close();
  }
}
